//Example usage: minimize energy E(x,y) = 2*x + 3*(y+1) + (x+1)*(y+2), where x,y \in {0,1}.

#include <stdio.h>
#include <vector>
#include "QPBO.h"

namespace{
std::vector< int > myqpbo( std::vector< double > unaries0, std::vector< double > unaries1, 
			   std::vector< int > binaries0, std::vector< int > binaries1, 
			   std::vector< double > binaries2, std::vector< double > binaries3, 
			   std::vector< double > binaries4, std::vector< double > binaries5 )
{
	typedef double REAL;
	QPBO<REAL>* q;

	int n_verts = (int) unaries0.size();
	int n_edges = (int) binaries0.size();
	assert( n_verts > 0 );
	assert( n_edges > 0 );
	
	q = new QPBO<REAL>(n_verts, n_edges); // max number of nodes & edges
	q->AddNode(n_verts); // add two nodes

	int i;
	for( i = 0; i < n_verts; i ++ ){
	  q->AddUnaryTerm(i, unaries0[i], unaries1[i]);
	}
	
	for( i = 0; i < n_edges; i ++ ){
	  q->AddPairwiseTerm( binaries0[i], binaries1[i], binaries2[i], binaries3[i], binaries4[i], binaries5[i]);
	}

	q->Solve();
	q->ComputeWeakPersistencies();

	std::vector< int > ret;
	
	for( i = 0; i < n_verts; i ++ )
	  ret.push_back( q->GetLabel(i) );
	
	delete q;
	
	return ret;
	
}
}