#!/usr/local/bin/python

# Synthetic experiments for graph mode problems

from vector import CGaussKernel,CLinearKernel,CRBFKernel
import numpy
from QPBO_Python import *
import pdb
from itertools import izip
import collections
import matplotlib.pyplot as plt
import sys

# hyper-parameters
etav = 0.005 # regularization in front of the clique-ness measure
thres = 0.85 # hyper-parameter for the adjacency matrices (reduces this number whenever QPBO returns -1)
omegas = 1.0 # hyper-parameter for the kernels
const = 1 # regularization in front of the original pairwise factor (might depend on the size of clique and time duration), for now, ignoring this constant so far works
debug_v = 1

# toy data
data = {}
numpy.random.seed(eval(sys.argv[1]))
data[0] = numpy.r_[1.0*numpy.random.randn(7, 2),2.0*numpy.random.randn(6, 2)+ numpy.array([-6, 3]),2.0*numpy.random.randn(5, 2)+ numpy.array([8, -3])] # time slice 0
data[1] = data[0] + 10.0*numpy.random.randn(18,2) # time slice 1 numpy.r_[1.0*numpy.random.randn(7, 2),2.0*numpy.random.randn(6, 2)+ numpy.array([-6, 3]),2.0*numpy.random.randn(5, 2)+ numpy.array([8, -3])]
data[2] = data[0] + 10.0*numpy.random.randn(18,2) # time slice 2 numpy.r_[1.0*numpy.random.randn(7, 2),2.0*numpy.random.randn(6, 2)+ numpy.array([-6, 3]),2.0*numpy.random.randn(5, 2)+ numpy.array([8, -3])]


#data[0] = numpy.r_[1.0*numpy.random.randn(7, 2),2.0*numpy.random.randn(6, 2)+ numpy.array([-6, 3]),2.0*numpy.random.randn(5, 2)+ numpy.array([8, -3])] # time slice 0
#data[1] = data[0] + 10.0*numpy.random.randn(18,2) # time slice 1 numpy.r_[1.0*numpy.random.randn(7, 2),2.0*numpy.random.randn(6, 2)+ numpy.array([-6, 3]),2.0*numpy.random.randn(5, 2)+ numpy.array([8, -3])]
#data[2] = data[0] + 1.0*numpy.random.randn(18,2) # time slice 2 numpy.r_[1.0*numpy.random.randn(7, 2),2.0*numpy.random.randn(6, 2)+ numpy.array([-6, 3]),2.0*numpy.random.randn(5, 2)+ numpy.array([8, -3])]
#data[3] = data[0] + 4.0*numpy.random.randn(18,2) # time slice 2 numpy.r_[1.0*numpy.random.randn(7, 2),2.0*numpy.random.randn(6, 2)+ numpy.array([-6, 3]),2.0*numpy.random.randn(5, 2)+ numpy.array([8, -3])]
#data[4] = data[0] + 8.0*numpy.random.randn(18,2) # time slice 2 numpy.r_[1.0*numpy.random.randn(7, 2),2.0*numpy.random.randn(6, 2)+ numpy.array([-6, 3]),2.0*numpy.random.randn(5, 2)+ numpy.array([8, -3])]

m = data[0].shape[0] # number of nodes 
ntime = len(data)

if debug_v:
	print "debugging flag is on \n"

# compute the (Gaussian RBF) kernel similarity matrices
dk = CRBFKernel()
K = {}
for i in xrange(ntime):
	dK = dk.Dot(data[i], data[i])
	omega_K = 1.0*omegas / numpy.median(dK.flatten())
	kernel_K = CGaussKernel(omega_K)
	K[i] = kernel_K.Dot(data[i], data[i]) 

Kavg = numpy.zeros((m,m)) # the average similarities across time slices
for i in xrange(ntime):
	Kavg = Kavg + K[i]

# build the graphs
G = collections.defaultdict(dict)

for itm in xrange(ntime):
	for i in xrange(m):
    		tt = K[itm][i,:] >= thres
    		tt[i] = False
    		ii = numpy.arange(len(K[itm][i,:]))[tt]
    		v = K[itm][i,:][tt]
    		G[itm][i] = dict(izip(ii,v))

unaries0 = VectorOfDouble()
unaries0.extend( numpy.zeros(m) ) # no unary term
unaries1 = VectorOfDouble()
unaries1.extend( numpy.zeros(m) ) # no unary term

# setting up the edges of the graph
binaries_id0 = VectorOfInt()
binaries_id1 = VectorOfInt()
for i in xrange(m):
	for j in xrange(i):
		binaries_id0.extend( [ i ] )
		binaries_id1.extend( [ j ] )
nedges = len(binaries_id0) # number of edges
print "# edges: ", nedges

# all edges have zero weight except for x_i=1 and x_j=1 configuration
binaries_weight_00 = VectorOfDouble()
binaries_weight_00.extend( numpy.zeros(nedges) )
binaries_weight_01 = VectorOfDouble()
binaries_weight_01.extend( numpy.zeros(nedges) )
binaries_weight_10 = VectorOfDouble()
binaries_weight_10.extend( numpy.zeros(nedges) )


dualvar = numpy.zeros(ntime) # initialize randomly the dual variables
for niter in xrange(2):
	# update indicator variables given the dual variables
	binaries_weight_11 = VectorOfDouble()
	for i in xrange(nedges):
		dissimilar = -Kavg[binaries_id0[i],binaries_id1[i]]
		beta_cliqueness = 0
		if debug_v:
			print "binaries_id0[i]: ",binaries_id0[i]
			print "binaries_id1[i]: ",binaries_id1[i]
		for itm in xrange(ntime):
			if not G[itm][binaries_id0[i]].has_key(binaries_id1[i]): #if no edge between binaries_id0[i] and binaries_id1[i]
				beta_cliqueness = beta_cliqueness + dualvar[itm]*1
		ellterm = (1.0/(4.0*etav))*numpy.sum(dualvar**2)
		if debug_v:
			print "beta_cliqueness: ",beta_cliqueness
			print "const*dissimilar: ",const*dissimilar
			print "reg on beta_cliqueness: ",ellterm
			print "\n"

		binaries_weight_11.extend( [const*dissimilar+beta_cliqueness] )
	
	# run QPBO
	qpbo_result = VectorOfInt()
	qpbo_result = myqpbo( unaries0, unaries1, binaries_id0, binaries_id1, binaries_weight_00, 		binaries_weight_01, binaries_weight_10, binaries_weight_11 )
	print numpy.array( qpbo_result )
	arrqpbo_result = numpy.array( qpbo_result )
	# update dual variables given the indicator variables
	for itm in xrange(ntime):
		tmpval = 0
		for i in xrange(nedges):
			if not G[itm][binaries_id0[i]].has_key(binaries_id1[i]): #if no edge between binaries_id0[i] and binaries_id1[i]
				tmpval = tmpval + arrqpbo_result[binaries_id0[i]]*arrqpbo_result[binaries_id1[i]]
		dualvar[itm] = 2.0*etav*tmpval
	print "dual variables: ", dualvar


for itm in xrange(ntime):
	plt.figure(itm)
	plt.scatter(data[itm][:,0],data[itm][:,1],s=100)
	plt.hold(True)
	plt.scatter(data[itm][0:7,0],data[itm][0:7,1],c='r',s=100) # the real ''family''
	plt.hold(True)
	cliques = numpy.arange(m)[numpy.array(qpbo_result)==1]
	for i in xrange(sum(numpy.array(qpbo_result))):
		for j in range(i,sum(numpy.array(qpbo_result))):
			plt.plot([data[itm][cliques[i],0],data[itm][cliques[j],0]],[data[itm][cliques[i],1],data[itm][cliques[j],1]])
			plt.hold(True)
	plt.hold(False)
	plt.xlabel('x-coordinate')
	plt.ylabel('y-coordinate')
	plt.grid(True)
plt.show()


